import inspect

from graceful.validators import min_validator, max_validator


class BaseField:
    """Base field class for subclassing.

    To create new field type subclass :any:`BaseField` and implement following
    methods:

    * ``from_representation()``: converts representation (used in
      request/response body) to internal value.

    * ``to_representation()``: converts internal value to representation
      that will be used in response body.


    Args:
        details (str): human readable description of field (it will be used
            for describing resource on OPTIONS requests).

        label (str): human readable label of a field (it will be used for
            describing resource on OPTIONS requests).

            *Note: it is recommended to use field names that are
            self-explanatory intead of relying on field labels.*

        source (str): name of internal object key/attribute that will be
            passed to field on ``.to_representation()`` call. Special ``'*'``
            value is allowed that will pass whole object to field when making
            representation. If not set then default source will
            be a field name used as a serializer's attribute.

        validators (list): list of validator callables.

        many (bool): set to True if field is in fact a list of given type
           objects.

        read_only (bool): True if field is read-only and cannot be set/modified
            via POST, PUT, or PATCH requests.

        write_only (bool): True if field is write-only and cannot be retrieved
            via GET requests.

        allow_null (bool): True if field can have intentional `null` values
            which will be interpreted as `None` afterwards.

            .. versionadded:: 0.5.0

    Example:

    .. code-block:: python

        class BoolField(BaseField):
            def from_representation(self, data):
                if data in {'true', 'True', 'yes', '1', 'Y'}:
                    return True:
                elif data in {'false', 'False', 'no', '0', 'N'}:
                    return False:
                else:
                    raise ValueError(
                        "{data} is not valid boolean field".format(
                            data=data
                        )
                    )

            def to_representation(self, value):
                return ["True", "False"][value]

    """

    #: Two-tuple ``(label, url)`` pointing to represented type specification
    #: (for documentation).
    spec = None

    #: String label of represented type (for documentation).
    type = None

    def __init__(
            self,
            details,
            label=None,
            source=None,
            validators=None,
            many=False,
            read_only=False,
            write_only=False,
            allow_null=False,
    ):
        """Initialize field definition."""
        self.label = label
        self.source = source
        self.details = details
        self.validators = validators or []
        self.many = many
        self.read_only = read_only
        self.write_only = write_only
        self.allow_null = allow_null

        if self.write_only and self.read_only:
            raise ValueError(
                "Field cannot be read-only and write-only at the same time."
            )

    def from_representation(self, data):
        """Convert representation value to internal value.

        Note:
             This is method handler stub and should be redifined in the
             ``BaseField`` subclasses.

        """
        raise NotImplementedError(
            "{cls}.from_representation() method not implemented".format(
                cls=self.__class__.__name__
            )
        )

    def to_representation(self, value):
        """Convert representation value to internal value.

        Note:
             This is method handler stub and should be redifined in the
             ``BaseField`` subclasses.

        """
        raise NotImplementedError(
            "{cls}.to_representation() method not implemented".format(
                cls=self.__class__.__name__
            )
        )

    def describe(self, **kwargs):
        """
        Describe this field instance for purpose of self-documentation.

        Args:
            kwargs (dict): dictionary of additional description items for
                extending default description

        Returns:
            dict: dictionary of description items

        Suggested way for overriding description fields or extending it with
        additional items is calling super class method with new/overriden
        fields passed as keyword arguments like following:

        .. code-block:: python

            class DummyField(BaseField):
               def description(self, **kwargs):
                   super().describe(is_dummy=True, **kwargs)

        """
        description = {
            'label': self.label,
            'details': inspect.cleandoc(self.details),
            'type': "list of {}".format(self.type) if self.many else self.type,
            'spec': self.spec,
            'read_only': self.read_only,
            'write_only': self.write_only,
            'allow_null': self.allow_null,
        }
        description.update(kwargs)
        return description

    def validate(self, value):
        """Perform validation on value by running all field validators.

        Single validator is a callable that accepts one positional argument
        and raises :any:`ValidationError` when validation fails.

        Error message included in exception will be included in http error
        response

        Args:
            value: internal value to validate

        Returns:
            None

        Note:
            Concept of validation for fields is understood here as a process
            of checking if data of valid type (successfully parsed/processed by
            ``.from_representation`` handler) does meet some other constraints
            (lenght, bounds, uniqueness, etc). So this method is always called
            with result of ``.from_representation()`` passed as its argument.

        """
        for validator in self.validators:
            validator(value)


class RawField(BaseField):
    """Represents raw field subtype.

    Any value from resource object will be returned as is without any
    conversion and no control over serialized value type is provided. Can be
    used only with very simple data types like int, float, str etc. but can
    eventually cause problems if value provided in representation has type
    that is not accepted in application.

    Effect of using this can differ between various content-types.

    """

    type = 'raw'

    def from_representation(self, data):
        """Return representation value as-is (note: content-type dependent)."""
        return data

    def to_representation(self, value):
        """Return internal value as-is (note: content-type dependent)."""
        return value


class StringField(BaseField):
    """Represents string field subtype without any extensive validation."""

    type = 'string'

    def from_representation(self, data):
        """Convert representation value to ``str``."""
        return str(data)

    def to_representation(self, value):
        """Convert representation value to ``str``."""
        return str(value)


class BoolField(BaseField):
    """Represents boolean type of field.

    By default accepts a wide range of incoming True/False representations:

    * False: ``['False', 'false', 'FALSE', 'F', 'f', '0', 0, 0.0, False]``
    * True: ``['True', 'true', 'TRUE', 'T', 't', '1', 1, True]``

    By default, the outup representations of internal object's value are
    Python's False/True values that will be later serialized to form that
    is native for content-type of use.

    This behavior can be changed using ``representations`` field argument.
    Note that when using ``representations`` parameter you need to make
    strict decision and there is no ability to accept multiple options for
    true/false representations. Anyway, it is reccomended approach to
    strictly define these values.

    Args:
        representations (tuple): two-tuple with representations for
            (False, True) values, that will be used instead of default values

    """

    type = 'bool'

    _TRUE_VALUES = {'True', 'true', 'TRUE', 'T', 't', '1', 1, True}
    _FALSE_VALUES = {'False', 'false', 'FALSE', 'F', 'f', '0', 0, 0.0, False}

    _DEFAULT_REPRESENTATIONS = (False, True)

    def __init__(
            self,
            details,
            representations=None,
            **kwargs
    ):
        """Initialize field definition and set preffered representations."""
        super().__init__(details, **kwargs)

        if representations:
            # could not resist...
            self._FALSE_VALUES = {representations[False]}
            self._TRUE_VALUES = {representations[True]}

        self.representations = representations or self._DEFAULT_REPRESENTATIONS

    def from_representation(self, data):
        """Convert representation value to ``bool`` if it has expected form."""
        if data in self._TRUE_VALUES:
            return True
        elif data in self._FALSE_VALUES:
            return False
        else:
            raise ValueError(
                "{type} type value must be one of {values}".format(
                    type=self.type,
                    values=self._TRUE_VALUES.union(self._FALSE_VALUES)
                )
            )

    def to_representation(self, value):
        """Convert internal boolean value to one of defined representations."""
        return self.representations[value]


class IntField(BaseField):
    """Represents integer type of field.

    Field of this type accepts both integers and strings as an incoming
    integer representation and always returns int as a representation of
    internal objects's value that will be later serialized to form that is
    native for content-type of use.

    This field accepts optional arguments that simply add new `max` and `min`
    value validation.

    Args:
        max_value (int): optional max value for validation
        min_value (int): optional min value for validation

    """

    type = 'int'

    def __init__(
            self,
            details,
            max_value=None,
            min_value=None,
            **kwargs
    ):
        """Initialize field definition and attach default validators."""
        super().__init__(details, **kwargs)

        self.max_value = max_value
        self.min_value = min_value

        if max_value is not None:
            self.validators.append(max_validator(max_value))
        if min_value is not None:
            self.validators.append(min_validator(min_value))

    def to_representation(self, value):
        """Convert internal value to ``int``."""
        return int(value)

    def from_representation(self, data):
        """Convert representation value to ``int``."""
        return int(data)


class FloatField(BaseField):
    """Represents float type of field.

    Accepts both floats and strings as an incoming float number
    representation and always returns float as a representation of internal
    objects's value that will be later serialized to form that is native for
    content-type of use.

    This field accepts optional arguments that simply add new `max` and `min`
    value validation.

    Args:
        max_value (int): optional max value for validation
        min_value (int): optional min value for validation

    """

    type = 'float'

    def __init__(
            self,
            details,
            max_value=None,
            min_value=None,
            **kwargs
    ):
        """Initialize field definition and attach default validators."""
        super().__init__(details, **kwargs)

        self.max_value = max_value
        self.min_value = min_value

        if max_value is not None:
            self.validators.append(max_validator(max_value))
        if min_value is not None:
            self.validators.append(min_validator(min_value))

    def to_representation(self, value):
        """Convert internal value to ``float``."""
        return float(value)

    def from_representation(self, data):
        """Convert representation value to ``float``."""
        return float(data)



import json
import inspect
from collections import OrderedDict
from warnings import warn

from falcon import errors
import falcon
from mimeparse import parse_mime_type

from graceful.parameters import BaseParam, IntParam
from graceful.errors import DeserializationError, ValidationError


class MetaResource(type):
    """Metaclass for handling parametrization with parameter objects."""

    _params_storage_key = '_params'

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        """Prepare class namespace in a way that ensures order of attributes.

        This needs to be an ``OrderedDict`` instance so ``_get_params()``
        method can construct params storage that preserves the same order of
        parameters as defined in code.

        Args:
            bases: all base classes of created resource class
            namespace (dict): namespace as dictionary of attributes

        """
        return OrderedDict()

    @classmethod
    def _get_params(mcs, bases, namespace):
        """Create params dictionary to be used in resource class namespace.

        Pop all parameter objects from attributes dict (namespace)
        and store them under _params_storage_key atrribute.
        Also collect all params from base classes in order that ensures
        params can be overriden.

        Args:
            bases: all base classes of created resource class
            namespace (dict): namespace as dictionary of attributes

        """
        params = [
            (name, namespace.pop(name))
            for name, attribute
            in list(namespace.items())
            if isinstance(attribute, BaseParam)
        ]

        for base in reversed(bases):
            if hasattr(base, mcs._params_storage_key):
                params = list(
                    getattr(base, mcs._params_storage_key).items()
                ) + params

        return OrderedDict(params)

    def __new__(mcs, name, bases, namespace, **kwargs):
        """Create new class object instance and alter its namespace."""
        namespace[mcs._params_storage_key] = mcs._get_params(bases, namespace)
        return super().__new__(
            # note: there is no need preserve order in namespace anymore so
            # we convert it explicitly to dict
            mcs, name, bases, dict(namespace),
        )

    def __init__(cls, name, bases, namespace, **kwargs):
        """Perform some additional class object initialization."""
        super().__init__(name, bases, namespace)

        try:
            # note: attribute is stored only if with_context keyword
            #       argument is not specified explicitly. This way
            #       we are able to guess if proper warning need to be
            #       displayed to the user.
            # future: remove in 1.x
            cls._with_context = kwargs.pop('with_context')
        except KeyError:
            pass


class BaseResource(metaclass=MetaResource):
    """Base resouce class with core param and response functionality.

    This base class handles resource responses, parameter deserialization,
    and validation of request included representations if serializer is
    defined.

    All custom resource classes based on ``BaseResource`` accept additional
    ``with_context`` keyword argument:


    .. code-block:: python

        class MyResource(BaseResource, with_context=True):
            ...

    The ``with_context`` argument tells if resource modification methods
    (metods injected with mixins - list/create/update/etc.) should accept
    the ``context`` argument in their signatures. For more details
    see :ref:`guide-context-aware-resources` section of documentation. The
    default value for ``with_context`` class keyword argument is ``False``.

    .. versionchanged:: 0.3.0
        Added the ``with_context`` keyword argument.

    """

    indent = IntParam(
        """
        JSON output indentation. Set to 0 if output should not be formated.
        """,
        default='0'
    )

    #: Instance of serializer class used to serialize/deserialize and
    #: validate resource representations.
    serializer = None

    def __new__(cls, *args, **kwargs):
        """Do some sanity checks before resource instance initialization."""
        instance = super().__new__(cls)

        if not hasattr(instance, '_with_context'):
            # note: warnings is displayed only if user did not specify
            #       explicitly that he want's his resource to not accept
            #       the context keyword argument.
            # future: remove in 1.x
            warn(
                """
                Class {} was defined without the 'with_context' keyword
                argument. This means that its resource manipulation
                methods (list/retrieve/create etc.) won't receive context
                keyword argument.

                This behaviour will change in 1.x version of graceful
                (please refer to documentation).
                """.format(cls),
                FutureWarning
            )
        return instance

    @property
    def params(self):
        """Return dictionary of parameter definition objects."""
        return getattr(self, self.__class__._params_storage_key)

    def make_body(self, resp, params, meta, content):
        """Construct response body in ``resp`` object using JSON serialization.

        Args:
            resp (falcon.Response): response object where to include
                serialized body
            params (dict): dictionary of parsed parameters
            meta (dict): dictionary of metadata to be included in 'meta'
                section of response
            content (dict): dictionary of response content (resource
                representation) to be included in 'content' section of response

        Returns:
            None

        """
        response = {
            'meta': meta,
            'content': content
        }
        resp.content_type = 'application/json'
        resp.body = json.dumps(
            response,
            indent=params['indent'] or None if 'indent' in params else None
        )

    def allowed_methods(self):
        """Return list of allowed HTTP methods on this resource.

        This is only for purpose of making resource description.

        Returns:
            list: list of allowed HTTP method names (uppercase)

        """
        return [
            method
            for method, allowed in (
                ('GET', hasattr(self, 'on_get')),
                ('POST', hasattr(self, 'on_post')),
                ('PUT', hasattr(self, 'on_put')),
                ('PATCH', hasattr(self, 'on_patch')),
                ('DELETE', hasattr(self, 'on_delete')),
                ('HEAD', hasattr(self, 'on_head')),
                ('OPTIONS', hasattr(self, 'on_options')),
            ) if allowed
        ]

    def describe(self, req=None, resp=None, **kwargs):
        """Describe API resource using resource introspection.

        Additional description on derrived resource class can be added using
        keyword arguments and calling ``super().decribe()`` method call
        like following:

        .. code-block:: python

             class SomeResource(BaseResource):
                 def describe(req, resp, **kwargs):
                     return super().describe(
                         req, resp, type='list', **kwargs
                      )

        Args:
            req (falcon.Request): request object
            resp (falcon.Response): response object
            kwargs (dict): dictionary of values created from resource url
                template

        Returns:
            dict: dictionary with resource descritpion information

        .. versionchanged:: 0.2.0
           The `req` and `resp` parameters became optional to ease the
           implementation of application-level documentation generators.
        """
        description = {
            'params': OrderedDict([
                (name, param.describe())
                for name, param in self.params.items()
            ]),
            'details':
                inspect.cleandoc(
                    self.__class__.__doc__ or
                    "This resource does not have description yet"
                ),
            'name': self.__class__.__name__,
            'methods': self.allowed_methods()
        }
        # note: add path to resource description only if request object was
        #       provided in order to make auto-documentation engines simpler
        if req:
            description['path'] = req.path

        description.update(**kwargs)
        return description

    def on_options(self, req, resp, **kwargs):
        """Respond with JSON formatted resource description on OPTIONS request.

        Args:
            req (falcon.Request): Optional request object. Defaults to None.
            resp (falcon.Response): Optional response object. Defaults to None.
            kwargs (dict): Dictionary of values created by falcon from
                resource uri template.

        Returns:
            None


        .. versionchanged:: 0.2.0
           Default ``OPTIONS`` responses include ``Allow`` header with list of
           allowed HTTP methods.
        """
        resp.set_header('Allow', ', '.join(self.allowed_methods()))
        resp.body = json.dumps(self.describe(req, resp))
        resp.content_type = 'application/json'

    def require_params(self, req):
        """Require all defined parameters from request query string.

        Raises ``falcon.errors.HTTPMissingParam`` exception if any of required
        parameters is missing and ``falcon.errors.HTTPInvalidParam`` if any
        of parameters could not be understood (wrong format).

        Args:
            req (falcon.Request): request object

        """
        params = {}

        for name, param in self.params.items():
            if name not in req.params and param.required:
                # we could simply raise with this single param or use get_param
                # with required=True parameter but for client convenience
                # we prefer to list all missing params that are required
                missing = set(
                    p for p in self.params
                    if self.params[p].required
                ) - set(req.params.keys())

                raise errors.HTTPMissingParam(", ".join(missing))

            elif name in req.params or param.default:
                # Note: lack of key in req.params means it was not specified
                # so unless there is default value it will not be included in
                # output params dict.
                # This way we have explicit information that param was
                # not specified. Using None would not be as good because param
                # class can also return None from `.value()` method as a valid
                # translated value.
                try:
                    if param.many:
                        # params with "many" enabled need special care
                        values = req.get_param_as_list(
                            # note: falcon allows to pass value handler using
                            #       `transform` param so we do not need to
                            #       iterate through list manually
                            name, param.validated_value
                        ) or [
                            param.default and
                            param.validated_value(param.default)
                        ]
                        params[name] = param.container(values)
                    else:
                        # note that if many==False and query parameter
                        # occurs multiple times in qs then it is
                        # **unspecified** which one will be used. See:
                        # http://falcon.readthedocs.org/en/latest/api/request_and_response.html#falcon.Request.get_param  # noqa
                        params[name] = param.validated_value(
                            req.get_param(name, default=param.default)
                        )

                except ValidationError as err:
                    # ValidationError allows to easily translate itself to
                    # to falcon's HTTPInvalidParam (Bad Request HTTP response)
                    raise err.as_invalid_param(name)

                except ValueError as err:
                    # Other parsing issues are expected to raise ValueError
                    raise errors.HTTPInvalidParam(str(err), name)

        return params

    def require_meta_and_content(self, content_handler, params, **kwargs):
        """Require 'meta' and 'content' dictionaries using proper hander.

        Args:
            content_handler (callable): function that accepts
                ``params, meta, **kwargs`` argument and returns dictionary
                for ``content`` response section
            params (dict): dictionary of parsed resource parameters
            kwargs (dict): dictionary of values created from resource url
                template

        Returns:
            tuple (meta, content): two-tuple with dictionaries of ``meta`` and
                ``content`` response sections

        """
        meta = {
            'params': params
        }
        content = content_handler(params, meta, **kwargs)
        meta['params'] = params
        return meta, content

    def require_representation(self, req):
        """Require raw representation dictionary from falcon request object.

        This does not perform any field parsing or validation but only uses
        allowed content-encoding handler to decode content body.

        Note:
            Currently only JSON is allowed as content type.

        Args:
            req (falcon.Request): request object

        Returns:
            dict: raw dictionary of representation supplied in request body

        """
        try:
            type_, subtype, _ = parse_mime_type(req.content_type)
            content_type = '/'.join((type_, subtype))
        except:
            raise falcon.HTTPUnsupportedMediaType(
                description="Invalid Content-Type header: {}".format(
                    req.content_type
                )
            )

        if content_type == 'application/json':
            body = req.stream.read()
            return json.loads(body.decode('utf-8'))
        else:
            raise falcon.HTTPUnsupportedMediaType(
                description="only JSON supported, got: {}".format(content_type)
            )

    def require_validated(self, req, partial=False, bulk=False):
        """Require fully validated internal object dictionary.

        Internal object dictionary creation is based on content-decoded
        representation retrieved from request body. Internal object validation
        is performed using resource serializer.

        Args:
            req (falcon.Request): request object
            partial (bool): set to True if partially complete representation
                is accepted (e.g. for patching instead of full update). Missing
                fields in representation will be skiped.
            bulk (bool): set to True if request payload represents multiple
                resources instead of single one.

        Returns:
            dict: dictionary of fields and values representing internal object.
                Each value is a result of ``field.from_representation`` call.

        """
        representations = [
            self.require_representation(req)
        ] if not bulk else self.require_representation(req)

        if bulk and not isinstance(representations, list):
            raise ValidationError(
                "Request payload should represent a list of resources."
            ).as_bad_request()

        object_dicts = []

        try:
            for representation in representations:
                object_dict = self.serializer.from_representation(
                    representation
                )
                self.serializer.validate(object_dict, partial)
                object_dicts.append(object_dict)

        except DeserializationError as err:
            # when working on Resource we know that we can finally raise
            # bad request exceptions
            raise err.as_bad_request()

        except ValidationError as err:
            # ValidationError is a suggested way to validate whole resource
            # so we also are prepared to catch it
            raise err.as_bad_request()

        return object_dicts if bulk else object_dicts[0]



# -*- coding: utf-8 -*-
from functools import wraps
from falcon import hooks, HTTPUnauthorized
import falcon.version

# todo: consider moving to `compat` module if we have to use more compat code
FALCON_VERSION = tuple(map(int, falcon.version.__version__.split('.')))


def _before(hook):
    """Metadata preserving version of ``falcon.hooks.before``.

    Args:
        hook: actual hook version.

    Returns: falcon hook decorator.
    """
    return wraps(hook)(hooks.before(hook))


@_before
def authentication_required(req, resp, resource, uri_kwargs):
    """Ensure that user is authenticated otherwise return ``401 Unauthorized``.

    If request fails to authenticate this authorization hook will also
    include list of ``WWW-Athenticate`` challenges.

    Args:
        req (falcon.Request): the request object.
        resp (falcon.Response): the response object.
        resource (object): the resource object.
        uri_kwargs (dict): keyword arguments from the URI template.

    .. versionadded:: 0.4.0
    """
    if 'user' not in req.context:
        args = ["Unauthorized", "This resource requires authentication"]

        # compat: falcon >= 1.0.0 requires the list of challenges
        if FALCON_VERSION >= (1, 0, 0):
            args.append(req.context.get('challenges', []))

        raise HTTPUnauthorized(*args)



# -*- coding: utf-8 -*-
import json
import base64
import binascii
import re
import abc

try:
    from functools import singledispatch
except ImportError:  # pragma: nocover
    # future: remove when dropping support for Python 3.3
    # compat: backport of singledispatch module introduced in Python 3.4
    from singledispatch import singledispatch

from falcon import HTTPMissingHeader, HTTPBadRequest


class BaseUserStorage(metaclass=abc.ABCMeta):
    """Base user storage class that defines required API for user storages.

    All built-in graceful authentication middleware classes expect user storage
    to have compatible API. Custom authentication middlewares do not need
    to use storages.

    .. versionadded:: 0.4.0
    """

    @abc.abstractmethod
    def get_user(
        self, identified_with, identifier, req, resp, resource, uri_kwargs
    ):
        """Get user from the storage.

        Args:
            identified_with (str): instance of the authentication middleware
                that provided the ``identifier`` value.
            identifier (str): string that identifies the user (it is specific
                for every authentication middleware implementation).
            req (falcon.Request): the request object.
            resp (falcon.Response): the response object.
            resource (object): the resource object.
            uri_kwargs (dict): keyword arguments from the URI template.

        Returns:
            the deserialized user object. Preferably a ``dict`` but it is
            application-specific.
        """
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def __subclasshook__(cls, klass):
        """Verify implicit class interface."""
        if cls is BaseUserStorage:
            if any("get_user" in B.__dict__ for B in klass.__mro__):
                return True
        return NotImplemented


class DummyUserStorage(BaseUserStorage):
    """A dummy storage that never returns users or returns specified default.

    This storage is part of :any:`Anonymous` authentication middleware.
    It may also be useful for testing purposes or to disable specific
    authentication middlewares through app configuration.

    Args:
        user: User object to return. Defaults to ``None`` (will never
        authenticate).

    .. versionadded:: 0.4.0
    """

    def __init__(self, user=None):
        """Initialize dummy storage."""
        self.user = user

    def get_user(
        self, identified_with, identifier, req, resp, resource, uri_kwargs
    ):
        """Return default user object."""
        return self.user


class IPRangeWhitelistStorage(BaseUserStorage):
    """Simple storage dedicated for :any:`XForwardedFor` authentication.

    This storage expects that authentication middleware return client address
    from its ``identify()`` method. For example usage see :any:`XForwardedFor`.
    Because it is IP range whitelist this storage it cannot distinguish
    different users' IP and always returns default user object. If you want to
    identify different users by their IP see :any:`KeyValueUserStorage`.

    Args:
        ip_range: Any object that supports ``in`` operator (i.e. implements the
            ``__cointains__`` method). The ``__contains__`` method should
            return ``True`` if identifier falls into specified whitelist.
            Tip: use ``iptools``.
        user: Default user object to return on successful authentication.

    .. versionadded:: 0.4.0
    """

    def __init__(self, ip_range, user):
        """Initialize IP whitelist storage."""
        self.ip_range = ip_range
        self.user = user

    def get_user(
        self, identified_with, identifier, req, resp, resource, uri_kwargs
    ):
        """Return default user object.

        .. note::
            This implementation expects that ``identifier`` is an user address.
        """
        if identifier in self.ip_range:
            return self.user


class KeyValueUserStorage(BaseUserStorage):
    """Basic user storage using any key-value store as authentication backend.

    Client identities are stored as string under keys matching following
    template::

        <key_prefix>:<identified_with>:<identifier>

    Where:

    * ``<key_prefix>`` is the configured key prefix (same as the initialization
      argument),
    * ``<identified_with>`` is the name of authentication middleware that
      provided user identifier,
    * ``<identifier>`` is the identifier object that identifies the user.

    Note that this key scheme will work only for middlewares that return
    identifiers as single string objects. Also the ``<identifier>`` part
    of key template is a plain text value of without any hashing algorithm
    applied. It may not be secure enough to store user secrets that way.

    If you want to use this storage with middleware that uses more complex
    identifier format/objects (e.g. the :any:`Basic` class) you will have
    to register own identifier format in the :any:`hash_identifier` method.
    For details see the :any:`hash_identifier` method docstring or the
    :ref:`practical example <auth-practical-example>` section of the
    documentation.

    Args:
        kv_store: Key-value store client instance (e.g. Redis client object).
            The ``kv_store`` must provide at least two methods: ``get(key)``
            and ``set(key, value)``. The arguments and return values of these
            methods must be strings.
        key_prefix: key prefix used to store client identities.
        serialization: serialization object/module that uses the
            ``dumps()``/``loads()`` protocol. Defaults to ``json``.

    .. versionadded:: 0.4.0
    """

    def __init__(self, kv_store, key_prefix='users', serialization=None):
        """Initialize kv_store user storage."""
        self.kv_store = kv_store
        self.key_prefix = key_prefix
        self.serialization = serialization or json

    def _get_storage_key(self, identified_with, identifier):
        """Get key string for given user identifier in consistent manner."""
        return ':'.join((
            self.key_prefix, identified_with.name,
            self.hash_identifier(identified_with, identifier),
        ))

    @staticmethod
    @singledispatch
    def hash_identifier(identified_with, identifier):
        """Create hash from identifier to be used as a part of user lookup.

        This method is a ``singledispatch`` function. It allows to register
        new implementations for specific authentication middleware classes:

        .. code-block:: python

            from hashlib import sha1

            from graceful.authentication import KeyValueUserStorage, Basic

            @KeyValueUserStorage.hash_identifier.register(Basic)
            def _(identified_with, identifier):
                return ":".join((
                    identifier[0],
                    sha1(identifier[1].encode()).hexdigest(),
                ))

        Args:
            identified_with (str): name of the authentication middleware used
                to identify the user.
            identifier (str): user identifier string

        Return:
            str: hashed identifier string
        """
        if isinstance(identifier, str):
            return identifier
        else:
            raise TypeError(
                "User storage does not support this kind of identifier"
            )

    def get_user(
        self, identified_with, identifier, req, resp, resource, uri_kwargs
    ):
        """Get user object for given identifier.

        Args:
            identified_with (object): authentication middleware used
                to identify the user.
            identifier: middleware specifix user identifier (string or tuple
                in case of all built in authentication middleware classes).

        Returns:
            dict: user object stored in Redis if it exists, otherwise ``None``
        """
        stored_value = self.kv_store.get(
            self._get_storage_key(identified_with, identifier)
        )
        if stored_value is not None:
            user = self.serialization.loads(stored_value.decode())
        else:
            user = None

        return user

    def register(self, identified_with, identifier, user):
        """Register new key for given client identifier.

        This is only a helper method that allows to register new
        user objects for client identities (keys, tokens, addresses etc.).

        Args:
            identified_with (object): authentication middleware used
                to identify the user.
            identifier (str): user identifier.
            user (str): user object to be stored in the backend.
        """
        self.kv_store.set(
            self._get_storage_key(identified_with, identifier),
            self.serialization.dumps(user).encode(),
        )


class BaseAuthenticationMiddleware:
    """Base class for all authentication middleware classes.

    Args:
        user_storage (BaseUserStorage): a storage object used to retrieve
            user object using their identifier lookup.
        name (str): custom name of the authentication middleware useful
            for handling custom user storage backends. Defaults to middleware
            class name.

    .. versionadded:: 0.4.0
    """

    #: challenge returned in WWW-Authenticate header on non authorized
    #: requests.
    challenge = None

    #: defines if Authentication middleware requires valid storage
    #: object to identify users
    only_with_storage = False

    def __init__(self, user_storage=None, name=None):
        """Initialize authentication middleware."""
        self.user_storage = user_storage
        self.name = (
            name if name else self.__class__.__name__
        )

        if (
            self.only_with_storage and
            not isinstance(self.user_storage, BaseUserStorage)
        ):
            raise ValueError(
                "{} authentication middleware requires valid storage. Got {}."
                "".format(self.__class__.__name__, self.user_storage)
            )

    def process_resource(self, req, resp, resource, uri_kwargs=None):
        """Process resource after routing to it.

        This is basic falcon middleware handler.

        Args:
            req (falcon.Request): request object
            resp (falcon.Response): response object
            resource (object): resource object matched by falcon router
            uri_kwargs (dict): additional keyword argument from uri template.
                For ``falcon<1.0.0`` this is always ``None``
        """
        if 'user' in req.context:
            return

        identifier = self.identify(req, resp, resource, uri_kwargs)
        user = self.try_storage(identifier, req, resp, resource, uri_kwargs)

        if user is not None:
            req.context['user'] = user

        # if did not succeed then we need to add this to list of available
        # challenges.
        elif self.challenge is not None:
            req.context.setdefault(
                'challenges', list()
            ).append(self.challenge)

    def identify(self, req, resp, resource, uri_kwargs):
        """Identify the user that made the request.

        Args:
            req (falcon.Request): request object
            resp (falcon.Response): response object
            resource (object): resource object matched by falcon router
            uri_kwargs (dict): additional keyword argument from uri template.
                For ``falcon<1.0.0`` this is always ``None``

        Returns:
            object: a user object (preferably a dictionary).
        """
        raise NotImplementedError  # pragma: nocover

    def try_storage(self, identifier, req, resp, resource, uri_kwargs):
        """Try to find user in configured user storage object.

        Args:
            identifier: User identifier.

        Returns:
            user object.
        """
        if identifier is None:
            user = None

        # note: if user_storage is defined, always use it in order to
        #       authenticate user.
        elif self.user_storage is not None:
            user = self.user_storage.get_user(
                self, identifier, req, resp, resource, uri_kwargs
            )

        # note: some authentication middleware classes may not require
        #       to be initialized with their own user_storage. In such
        #       case this will always authenticate with "syntetic user"
        #       if there is a valid indentity.
        elif self.user_storage is None and not self.only_with_storage:
            user = {
                'identified_with': self,
                'identifier': identifier
            }

        else:  # pragma: nocover
            # note: this should not happen if the base class is properly
            #       initialized. Still, user can skip super().__init__() call.
            user = None

        return user


class Basic(BaseAuthenticationMiddleware):
    """Authenticate user with Basic auth as specified by `RFC 7617`_.

    Token authentication takes form of ``Authorization`` header in the
    following form::

        Authorization: Basic <credentials>

    Whre `<credentials>` is base64 encoded username and password separated by
    single colon charactes (refer to official RFC). Usernames must not contain
    colon characters!

    If client fails to authenticate on protected endpoint the response will
    include following challenge::

        WWW-Authenticate: Basic realm="<realm>"

    Where ``<realm>`` is the value of configured authentication realm.

    This middleware **must** be configured with ``user_storage`` that provides
    access to database of client API keys and their identities. Additionally.
    the ``identifier`` received by user storage in the ``get_user()`` method
    is a decoded ``<username>:<password>`` string. If you need to apply any
    hash function before hitting database in your user storage handler, you
    should split it using followitg code::

        username, _, password = identifier.partition(":")

    Args:
        realm (str): name of the protected realm. This can be only alphanumeric
            string with spaces (see: the ``REALM_RE`` pattern).
        user_storage (BaseUserStorage): a storage object used to retrieve
            user object using their identifier lookup.
        name (str): custom name of the authentication middleware useful
            for handling custom user storage backends. Defaults to middleware
            class name.

    .. versionadded:: 0.4.0

    .. _RFC 7617: https://tools.ietf.org/html/rfc7616
    """

    only_with_storage = True

    #: regular expression used to validate configured realm
    REALM_RE = re.compile(r"^[\w ]+$")

    def __init__(self, user_storage=None, name=None, realm="api"):
        """Initialize middleware and validate realm string."""
        if not self.REALM_RE.match(realm):
            raise ValueError(
                "realm argument should match '{}' regular expression"
                "".format(self.REALM_RE.pattern)
            )

        self.challenge = "Basic realm={}".format(realm)
        super(Basic, self).__init__(user_storage, name)

    def identify(self, req, resp, resource, uri_kwargs):
        """Identify user using Authenticate header with Basic auth."""
        header = req.get_header("Authorization", False)
        auth = header.split(" ") if header else None

        if auth is None or auth[0].lower() != 'basic':
            return None

        if len(auth) != 2:
            raise HTTPBadRequest(
                "Invalid Authorization header",
                "The Authorization header for Basic auth should be in form:\n"
                "Authorization: Basic <base64-user-pass>"
            )

        user_pass = auth[1]

        try:
            decoded = base64.b64decode(user_pass).decode()

        except (TypeError, UnicodeDecodeError, binascii.Error):
            raise HTTPBadRequest(
                "Invalid Authorization header",
                "Credentials for Basic auth not correctly base64 encoded."
            )

        username, _, password = decoded.partition(":")
        return username, password


class XAPIKey(BaseAuthenticationMiddleware):
    """Authenticate user with ``X-Api-Key`` header.

    The X-Api-Key authentication takes a form of ``X-Api-Key`` header in the
    following form::

        X-Api-Key: <key_value>

    Where ``<key_value>`` is a secret string known to both client and server.
    Example of valid header::

        X-Api-Key: 6fa459ea-ee8a-3ca4-894e-db77e160355e

    If client fails to authenticate on protected endpoint the response will
    include following challenge::

        WWW-Authenticate: X-Api-Key

    .. note::
        This method functionally equivalent to :any:`Token` and is included
        only to ease migration of old applications that could use such
        authentication method in past. If you're building new API and require
        only simple token-based authentication you should prefere
        :any:`Token` middleware.

    This middleware **must** be configured with ``user_storage`` that provides
    access to database of client API keys and their identities.

    .. versionadded:: 0.4.0
    """

    challenge = 'X-Api-Key'
    only_with_storage = True

    def identify(self, req, resp, resource, uri_kwargs):
        """Initialize X-Api-Key authentication middleware."""
        try:
            return req.get_header('X-Api-Key', True)
        except (KeyError, HTTPMissingHeader):
            pass


class Token(BaseAuthenticationMiddleware):
    """Authenticate user using Token authentication.

    Token authentication takes form of ``Authorization`` header::

        Authorization: Token <token_value>

    Where ``<token_value>`` is a secret string known to both client and server.
    Example of valid header::

        Authorization: Token 6fa459ea-ee8a-3ca4-894e-db77e160355e

    If client fails to authenticate on protected endpoint the response will
    include following challenge::

        WWW-Authenticate: Token

    This middleware **must** be configured with ``user_storage`` that provides
    access to database of client tokens and their identities.

    .. versionadded:: 0.4.0
    """

    challenge = 'Token'
    only_with_storage = True

    def identify(self, req, resp, resource, uri_kwargs):
        """Identify user using Authenticate header with Token auth."""
        header = req.get_header('Authorization', False)
        auth = header.split(' ') if header else None

        if auth is None or auth[0].lower() != 'token':
            return None

        if len(auth) != 2:
            raise HTTPBadRequest(
                "Invalid Authorization header",
                "The Authorization header for Token auth should be in form:\n"
                "Authorization: Token <token_value>"
            )

        return auth[1]


class XForwardedFor(BaseAuthenticationMiddleware):
    """Authenticate user with ``X-Forwarded-For`` header or remote address.

    Args:
        remote_address_fallback (bool): Use fallback to ``REMOTE_ADDR`` value
            from WSGI environment dictionary if ``X-Forwarded-For`` header is
            not available. Defaults to ``False``.


    This authentication middleware is usually used with the
    :any:`IPRangeWhitelistStorage` e.g:


    .. code-block:: python

        from iptools import IPRangeList
        import falcon

        from graceful import authentication

        IP_WHITELIST = IpRangeList(
            '127.0.0.1',
            # ...
        )

        auth_middleware = authentication.XForwardedFor(
            user_storage=authentication.IPWRangehitelistStorage(
                IP_WHITELIST, user={"username": "internal"}
            )
        )

        api = application = falcon.API(middleware=[auth_middleware])

    .. note::
        Using this middleware class is **highly unrecommended** if you
        are not able to ensure that contents of ``X-Forwarded-For`` header
        can be trusted. This requires proper reverse proxy and network
        configuration. It is also recommended to at least use the static
        :any:`IPRangeWhitelistStorage` as the user storage.

    .. versionadded:: 0.4.0
    """

    challenge = None
    only_with_storage = False

    def __init__(
        self, user_storage=None, name=None, remote_address_fallback=False
    ):
        """Initialize middleware and set default arguments."""
        super().__init__(user_storage, name)
        self.remote_address_fallback = remote_address_fallback

    def _get_client_address(self, req):
        """Get address from ``X-Forwarded-For`` header or use remote address.

        Remote address is used if the ``X-Forwarded-For`` header is not
        available. Note that this may not be safe to depend on both without
        proper authorization backend.

        Args:
            req (falcon.Request): falcon.Request object.

        Returns:
            str: client address.
        """
        try:
            forwarded_for = req.get_header('X-Forwarded-For', True)
            return forwarded_for.split(',')[0].strip()
        except (KeyError, HTTPMissingHeader):
            return (
                req.env.get('REMOTE_ADDR') if self.remote_address_fallback
                else None
            )

    def identify(self, req, resp, resource, uri_kwargs):
        """Identify client using his address."""
        return self._get_client_address(req)


class Anonymous(BaseAuthenticationMiddleware):
    """Dummy authentication middleware that authenticates every request.

    It makes every every request authenticated with default value of
    anonymous user. This authentication middleware may be used in order
    to simplify custom authorization code since it will ensure that
    every request context will have the ``'user'`` variable defined.

    .. note::
        This middleware will always add the default user to the request
        context if no other previous authentication middleware resolved.
        So if this middleware is used it makes no sense to:

        * Use the :any:`authentication_required` decorator.
        * Define any other authentication middleware after this one.

    Args:
        user: Default anonymous user object.

    .. versionadded:: 0.4.0
    """

    challenge = None
    only_with_storage = True

    def __init__(self, user):
        """Initialize anonymous authentication middleware."""
        # note: DummyUserStorage allows to always return the same user
        #       object that was passed as initialization value
        super().__init__(user_storage=DummyUserStorage(user))

    def identify(self, req, resp, resource, uri_kwargs):
        """Identify user with a dummy sentinel value."""
        # note: this is just a sentinel value to trigger successful
        #       lookup in the dummy user storage
        return ...
